"""CompletionObject class.
"""
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class CompletionObject(FoundationModelObject):
    """
    A class representing a completion response object.
    """

    @property
    def text(self):
        """
        Returns the text content from the chat completion API response.

        Returns:
            List[str]: The text content.
        """
        return [data['text'] for data in self.response['choices']]



"""CompletionChunkObject class.
"""
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class CompletionChunkObject(FoundationModelObject):
    """
    A class representing a chunk of completed text generated by the model, when `stream` is set True.
    """

    @property
    def text(self):
        """
        Returns the text content from the chat completion API response.

        Returns:
            str: The text content.
        """
        return self.response['choices'][0]['text']



"""Text Completion API resource.
"""
from typing import List, Optional, Union

from databricks_genai_inference.api.abstract.foundation_model_api_resource import (FoundationModelAPIInput,
                                                                                   FoundationModelAPIResource)
from databricks_genai_inference.api.objects.completion_chunk_object import CompletionChunkObject
from databricks_genai_inference.api.objects.completion_object import CompletionObject
from databricks_genai_inference.api.util import CompletionModel


class CompletionAPIInput(FoundationModelAPIInput):
    """
    A class representing the input parameters for the completion API.

    Attributes:
        prompt (Union[str, List[str]]): The prompt(s) to generate completions. Can be a string or a list of strings.
        user (Optional[str]): An id representing the user making the request.
        max_tokens (Optional[int]): The maximum number of tokens to generate.
        temperature (Optional[float]): The sampling temperature. Use higher value for more random outputs and lower value for more deterministic outputs. Must be between 0 and 2. Defaults to 1.0.
        top_p (Optional[float]): The probability threshold used for nucleus sampling. Must be between 0 and 1. Defaults to 1.
        top_k (Optional[int]): Defines the number of k most likely tokens to use for top-k-filtering. Set this value to 1 to make outputs deterministic.
        stream (Optional[bool]): If set to True, the API will stream the partial output as it’s generated as message chunks. Defaults to False.
        n (Optional[int]): The number of completion choices to return. Currently, only 1 choice is supported.
        suffix (Optional[str]): A string that is appended to the end of every completion. Defaults to an empty string.
        echo (Optional[bool]): If set to True, the API will echo the prompt in the completion. Defaults to False.
        error_behavior (Optional[str]): Error behavior when timeouts or context-length-exceeded errors happen. Two options: “truncate” (return as many tokens as possible) and “error” (return an error). Defaults to "error".
        stop (Optional[Union[str, List[str]]]): API will stop generating further tokens when any one of the sequences in stop are encountered.
        use_raw_prompt (Optional[bool]): If set to True, API will skip prompt template and use the raw prompt. Defaults to False.
    """
    prompt: Union[str, List[str]]
    user: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    stream: Optional[bool] = None
    n: Optional[int] = None
    suffix: Optional[str] = None
    echo: Optional[bool] = None
    error_behavior: Optional[str] = None
    stop: Optional[Union[str, List[str]]] = None
    use_raw_prompt: Optional[bool] = None


class Completion(FoundationModelAPIResource):
    """
    A class representing the text completion API resource.
    """

    SUPPORTED_MODEL_LIST = [model.value for model in CompletionModel.__members__.values()]
    model_input = CompletionAPIInput
    model_output = CompletionObject
    model_streaming_output = CompletionChunkObject



"""Utils for the API.
"""
from enum import Enum


class EmbeddingModel(Enum):
    """Supported embedding models.
    """
    BGE_LARGE_ENG = 'bge-large-en'


class CompletionModel(Enum):
    """Supported completion models.
    """
    MPT_7B_INSTRUCT = 'mpt-7b-instruct'
    # MPT_30B_INSTRUCT = 'mpt-30b-instruct' # not offered for prpr


class ChatCompletionModel(Enum):
    """Supported chat completion models.
    """
    LLAMA_2_70B_CHAT = 'llama-2-70b-chat'



"""Foundation Model API Resource.
"""
import json as json_lib
from typing import Optional

import requests
from databricks_cli.configure import provider
from pydantic import BaseModel, ConfigDict, ValidationError

from databricks_genai_inference.api.abstract.api_resource import APIResource
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject
from databricks_genai_inference.api.exception import FoundationModelAPIException


class FoundationModelAPIInput(BaseModel):
    """
    A class representing the input schema for a Foundation Model API request.

    Attributes:
        model (str): The name of the model to use.
        timeout (Optional[int]): The timeout for the API request.
    """
    model_config = ConfigDict(extra='forbid')

    model: str
    timeout: Optional[int] = None


class FoundationModelAPIResource(APIResource):
    """
    A class representing a foundation model API resource.

    Attributes:
        SUPPORTED_MODEL_LIST (list): A list of supported models.
        DEFAULT_TIMEOUT (int): The default timeout for API requests.
        model_input (FoundationModelAPIInput): The input schema for the API.
        model_output (FoundationModelObject): The output schema for the API.
        model_streaming_output (FoundationModelObject): The streaming output schema for the API.
        config (Config): The configuration object for the API.
    """

    SUPPORTED_MODEL_LIST = []
    DEFAULT_TIMEOUT = 60
    model_input = FoundationModelAPIInput
    model_output = FoundationModelObject
    model_streaming_output = FoundationModelObject

    @classmethod
    def create(cls, **kwargs):
        """
        Creates a new API response.

        Args:
        **kwargs: The keyword arguments for the API.

        Returns:
        The result of the API query.
        """
        api_input = cls._parse_and_validate_request(**kwargs)
        return cls._make_query(api_input)

    @classmethod
    def _parse_and_validate_request(cls, **kwargs) -> FoundationModelAPIInput:
        """
        Parses and validates the request parameters.

        Args:
            **kwargs: The request parameters.

        Returns:
            A FoundationModelAPIInput object.

        Raises:
            FoundationModelAPIException: If the request parameters are invalid.
        """
        try:
            api_input = cls.model_input(**kwargs)
            if api_input.model not in cls.SUPPORTED_MODEL_LIST:
                raise FoundationModelAPIException(
                    message=f"Invalid model name: {api_input.model}. Supported model list: {cls.SUPPORTED_MODEL_LIST}")
            return api_input
        except ValidationError as e:
            raise FoundationModelAPIException(message=str(e)) from e

    @classmethod
    def _make_query(cls, model_input: FoundationModelAPIInput):
        """
        Makes a query to the API.

        Args:
        model_input (FoundationModelAPIInput): The input for the API.

        Returns:
        The response of the API query.

        Raises:
        FoundationModelAPIException: If the API query fails.
        """
        url = f'{provider.get_config().host}/serving-endpoints/databricks-{model_input.model}/invocations'
        headers = {'authorization': f'Bearer {provider.get_config().token}', 'Content-Type': 'application/json'}
        json = model_input.model_dump(exclude_unset=True)
        timeout = json.pop("timeout", cls.DEFAULT_TIMEOUT)
        try:
            if model_input.model_dump().get("stream", False):
                return cls._get_streaming_response(url=url, headers=headers, json=json, timeout=timeout)
            else:
                return cls._get_non_streaming_response(url=url, headers=headers, json=json, timeout=timeout)
        except requests.exceptions.ReadTimeout as e:
            raise FoundationModelAPIException(message=f'API request timed out after {timeout} seconds') from e
        except requests.exceptions.ConnectionError as e:
            raise FoundationModelAPIException(message="API request failed with connection error") from e
        except FoundationModelAPIException as e:
            raise e

    @classmethod
    def _get_non_streaming_response(cls, url, headers, json, timeout):
        """
        Sends a request to the API and returns the non-streaming response.

        Args:
        url (str): The URL for the API.
        headers (dict): The headers for the API request.
        json (dict): The JSON data for the API request.
        timeout (int): The timeout for the API request.

        Raises:
        NotImplementedError: If the method is not implemented.
        """
        response = requests.post(url=url, headers=headers, json=json, timeout=timeout)
        if response.ok:
            try:
                return cls.model_output(response.json())
            except requests.JSONDecodeError as e:
                raise FoundationModelAPIException(response=response, url=url) from e
        else:
            raise FoundationModelAPIException(response=response, url=url)

    @classmethod
    def _get_streaming_response(cls, url, headers, json, timeout):
        """
        Sends a request to the API and returns the streaming response.

        Args:
        url (str): The URL for the API.
        headers (dict): The headers for the API request.
        json (dict): The JSON data for the API request.
        timeout (int): The timeout for the API request.

        Raises:
        NotImplementedError: If the method is not implemented.
        """
        with requests.post(url=url, headers=headers, json=json, timeout=timeout, stream=True) as response:
            if response.ok:
                try:
                    for line in response.iter_lines():
                        if line:
                            line = line.decode('utf-8')
                            if line.startswith("data: "):
                                line = line[len("data: "):]
                            if line == '[DONE]':
                                break
                            loaded_json = json_lib.loads(line)
                            if loaded_json:
                                yield cls.model_streaming_output(loaded_json)
                except json_lib.decoder.JSONDecodeError as e:
                    raise FoundationModelAPIException(response=response, message="JSONDecodeError", url=url) from e
            else:
                raise FoundationModelAPIException(response=response, url=url)



from abc import ABC, abstractmethod
from typing import Generator, Union

from pydantic import BaseModel

from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class APIResource(ABC):

    @classmethod
    @abstractmethod
    def create(cls, **kwargs) -> Union[FoundationModelObject, Generator[FoundationModelObject, None, None]]:
        pass

    @classmethod
    @abstractmethod
    def _parse_and_validate_request(cls, **kwargs) -> BaseModel:
        pass

    @classmethod
    @abstractmethod
    def _get_non_streaming_response(cls, url, headers, json, timeout) -> FoundationModelObject:
        pass

    @classmethod
    @abstractmethod
    def _get_streaming_response(cls, url, headers, json, timeout) -> Generator[FoundationModelObject, None, None]:
        pass



"""Foundation Model API Response Object.
"""
import json
from abc import ABC


class FoundationModelObject(ABC):
    """
    Abstract base class for all api response objects.
    """

    def __init__(self, response):
        """
        Initializes a new instance of the FoundationModelObject class.

        Args:
            response (dict): The response from the Foundation Model API.
        """
        self.response = response

    @property
    def json(self):
        """
        Gets the raw JSON representation of the response.

        Returns:
            dict: The raw JSON representation of the response.
        """
        return self.response

    @property
    def id(self):
        """
        Gets the ID of the model.

        Returns:
            str: The ID of the model.
        """
        return self.response['id']

    @property
    def model(self):
        """
        Gets the name of the model.

        Returns:
            str: The name of the model.
        """
        return self.response['model']

    @property
    def usage(self):
        """
        Gets the usage meta data.

        Returns:
            dict: The usage meta data.
        """
        return self.response['usage']

    def __str__(self):
        """
        Gets the string representation of the response.

        Returns:
            str: The string representation of the response.
        """
        return json.dumps(self.response, indent=2)



""" ChatSession class for multi-turn chat completion management.
"""
import copy

from databricks_genai_inference.api.chat_completion import ChatCompletion


class ChatSession():
    """
    A class representing a chat session with a chat completion model.

    Methods:
        reply(message: str) -> ChatCompletion: Sends a message to the chat model and returns its response.
        system_message() -> str: Returns the system message.
        last() -> str: Returns the last assistant message in the chat history.
        history() -> List: Returns the entire chat history.
        pretty_history() -> str: Returns a formatted string representing the chat history.
        count() -> int: Returns the number of chat rounds conducted so far
    """

    def __init__(self, model: str, system_message=None, **kwargs):
        """Args:
            model (str): The model name.
            system_message (str): The system message to guide the conversation. e.g. "You are a helpful assistant."
            **kwargs: Additional model parameters to pass to the chat completion API.
        """
        self.model = model
        self.parameters = kwargs
        self.chat_history = []
        self.system_message = system_message
        if self.parameters.get("stream", False):
            raise NotImplementedError(
                "You are setting stream=True, but streaming is not supported for ChatSession() yet.")
        if system_message is not None:
            self.chat_history = [{"role": "system", "content": system_message}]

    def reply(self, message: str):
        """
        Sends a message to the chat model and returns its response.

        Args:
            message (str): The message to send to the chat model.

        Returns:
            ChatCompletionObject: An object representing the response from the chat model.
        """
        self.chat_history.append({"role": "user", "content": message})
        response = ChatCompletion.create(model=self.model, messages=copy.deepcopy(self.chat_history), **self.parameters)
        self.chat_history.append({"role": "assistant", "content": response.message})
        return response

    @property
    def last(self):
        """
        Returns the last assistant message in the chat history.

        Returns:
            str: The last assistant message in the chat history.
        """
        if len(self.chat_history) <= 1:
            return ""
        return self.chat_history[-1]["content"]

    @property
    def history(self):
        """
        Returns the entire chat history.

        Returns:
            List: The entire chat history.
        """
        return self.chat_history

    @property
    def pretty_history(self):
        """
        Returns a formatted string representing the chat history.

        Returns:
            str: A formatted string representing the chat history.
        """
        return '\n' + '\n'.join([
            f'{self.chat_history[i]["role"]}: {self.chat_history[i]["content"]}' for i in range(len(self.chat_history))
        ])

    @property
    def count(self):
        """
        Returns the number of chat rounds conducted so far.

        Returns:
            int: The number of chat rounds conducted so far.
        """
        return int(len(self.chat_history) / 2)



"""ChatCompletionObject class.
"""
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class ChatCompletionObject(FoundationModelObject):
    """
    A class representing a chat completion response object.
    """

    @property
    def message(self):
        """
        Returns the message content from the chat completion API response.

        Returns:
            str: The message content.
        """
        return self.response['choices'][0]['message']['content']



"""ChatCompletionChunkObject class
"""
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class ChatCompletionChunkObject(FoundationModelObject):
    """
    A class representing a chunk of completed chat message generated by the model, when `stream` is set True.
    """

    @property
    def message(self):
        """
        Returns the message content from the chat completion API response.

        Returns:
            str: The message content.
        """
        return self.response['choices'][0]['delta']["content"]



"""Embedding API resource.
"""
from typing import List, Optional, Union

from databricks_genai_inference.api.abstract.foundation_model_api_resource import (FoundationModelAPIInput,
                                                                                   FoundationModelAPIResource)
from databricks_genai_inference.api.objects.embedding_object import EmbeddingObject
from databricks_genai_inference.api.util import EmbeddingModel


class EmbeddingAPIInput(FoundationModelAPIInput):
    """
    A class representing the input schema for the Embedding API.

    Attributes:
        input (Union[str, List[str]]): The input text to embed. Can be a string or a list of strings.
        instruction (Optional[str]): The task instruction. If not provided, only the input will be embedded.
        user (Optional[str]): An id representing the user making the request.
    """
    input: Union[str, List[str]]
    instruction: Optional[str] = None
    user: Optional[str] = None


class Embedding(FoundationModelAPIResource):
    """
    A class representing the embedding API resource.
    """
    SUPPORTED_MODEL_LIST = [model.value for model in EmbeddingModel.__members__.values()]
    model_input = EmbeddingAPIInput
    model_output = EmbeddingObject

    @classmethod
    def _get_streaming_response(cls, url, headers, json, timeout):
        raise NotImplementedError("Streaming is not supported for the Embedding API.")



"""EmbeddingObject class.
"""
from databricks_genai_inference.api.abstract.foundation_model_object import FoundationModelObject


class EmbeddingObject(FoundationModelObject):
    """
    A class representing an embedding response object.
    """

    @property
    def embeddings(self):
        """
        Returns the embedding content from the chat completion API response.

        Returns:
            List: The embedding content.
        """
        return [data['embedding'] for data in self.response['data']]



"""Databricks Generative AI Inference Package
"""
from databricks_genai_inference.api import (ChatCompletion, ChatCompletionChunkObject, ChatCompletionObject,
                                            ChatSession, Completion, CompletionChunkObject, CompletionObject, Embedding,
                                            EmbeddingObject, FoundationModelAPIException)

from .version import __version__

__all__ = [
    "ChatCompletion", "ChatSession", "Completion", "Embedding", "FoundationModelAPIException", "ChatCompletionObject",
    "ChatCompletionChunkObject", "CompletionObject", "CompletionChunkObject", "EmbeddingObject"
]



"""Exceptions for the foundation model api
"""
import logging
from http import HTTPStatus

import requests

logger = logging.getLogger(__name__)

DEFAULT_MESSAGE = 'Unknown Error'
DEFAULT_URL = None


class FoundationModelAPIException(Exception):
    """Exception raised when foundation model api requests fail

    Attributes:
        status (HTTPStatus): HTTP status code of the response
        message (str): Error message returned by the API
        url (str): URL of the API endpoint that was called
    """

    def __init__(self,
                 status: HTTPStatus = None,
                 message: str = DEFAULT_MESSAGE,
                 url: str = DEFAULT_URL,
                 response: requests.Response = None):
        self.status = status
        self.message = message
        self.url = url
        if response is not None:
            try:
                self.status = HTTPStatus(
                    response.status_code) if response.status_code != HTTPStatus.OK else HTTPStatus.INTERNAL_SERVER_ERROR
            except ValueError:
                logger.debug(f'Unknown status code {response.status_code}. Setting to 500')
                self.status = HTTPStatus.INTERNAL_SERVER_ERROR
            error = response.content.decode().strip()
            self.message = error if error else self.message

    def __str__(self) -> str:
        error_string = (f'\ncode: {self.status.value}' if self.status else '') + (
            f'\nreason: {self.message}' if self.message else '') + (f'\nurl: {self.url}' if self.url else '')
        return error_string



import io
import setuptools
from setuptools import setup

# pylint: disable-next=exec-used,consider-using-with
exec(open('databricks_genai_inference/version.py', 'r', encoding='utf-8').read())

install_requires = [
    'pyyaml>=5.4.1',
    'requests>=2.26.0,<3',
    'databricks-cli>=0.17.8',
    'databricks-sdk>=0.12.0',
    'pydantic>=2.4.2',
    'typing_extensions>=4.7.1',
]

extra_deps = {}

extra_deps['dev'] = [
    'build>=0.10.0',
    'isort>=5.9.3',
    'pre-commit>=2.17.0',
    'pylint>=2.12.2',
    'pyright==1.1.256',
    'pytest-cov>=4.0.0',
    'pytest-mock>=3.7.0',
    'pytest>=6.2.5',
    'radon>=5.1.0',
    'twine>=4.0.2',
    'toml>=0.10.2',
    'yapf>=0.33.0',
]

extra_deps['all'] = set(dep for deps in extra_deps.values() for dep in deps)

setup(
    name='databricks-genai-inference',
    version=__version__,  # type: ignore pylint: disable=undefined-variable
    author='Databricks',
    author_email='eng-genai-inference@databricks.com',
    description='Interact with the Databricks Foundation Model API from python',
    long_description=io.open("README.md", encoding="utf-8").read(),
    long_description_content_type='text/markdown',
    url='https://docs.databricks.com/en/machine-learning/foundation-models/query-foundation-model-apis.html',
    include_package_data=True,
    package_data={},
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=install_requires,
    extras_require=extra_deps,
    python_requires='>=3.9',
    ext_package='databricks_genai_inference',
)


from databricks_genai_inference.api.chat_completion import ChatCompletion
from databricks_genai_inference.api.chat_session import ChatSession
from databricks_genai_inference.api.completion import Completion
from databricks_genai_inference.api.embedding import Embedding
from databricks_genai_inference.api.exception import FoundationModelAPIException
from databricks_genai_inference.api.objects.chat_completion_chunk_object import ChatCompletionChunkObject
from databricks_genai_inference.api.objects.chat_completion_object import ChatCompletionObject
from databricks_genai_inference.api.objects.completion_chunk_object import CompletionChunkObject
from databricks_genai_inference.api.objects.completion_object import CompletionObject
from databricks_genai_inference.api.objects.embedding_object import EmbeddingObject



"""Chat Completion API resource.
"""
from typing import List, Optional, Union

from databricks_genai_inference.api.abstract.foundation_model_api_resource import (FoundationModelAPIInput,
                                                                                   FoundationModelAPIResource)
from databricks_genai_inference.api.objects.chat_completion_chunk_object import ChatCompletionChunkObject
from databricks_genai_inference.api.objects.chat_completion_object import ChatCompletionObject
from databricks_genai_inference.api.util import ChatCompletionModel


class ChatCompletionAPIInput(FoundationModelAPIInput):
    """
    A class representing the input schema for the Chat Completion API.

    Attributes:
        messages (List[dict]): A list of messages comprising the conversation so far. Each message is a dictionary with the following keys: `role` (str), `content` (str)
        user (Optional[str]): An id representing the user making the request.
        max_tokens (Optional[int]): The maximum number of tokens to generate.
        temperature (Optional[float]): The sampling temperature. Use higher value for more random outputs and lower value for more deterministic outputs. Must be between 0 and 2. Defaults to 1.0.
        top_p (Optional[float]): The probability threshold used for nucleus sampling. Must be between 0 and 1. Defaults to 1.
        top_k (Optional[int]): Defines the number of k most likely tokens to use for top-k-filtering. Set this value to 1 to make outputs deterministic.
        stream (Optional[bool]): If set to True, the API will stream the partial output as it’s generated as message chunks. Defaults to False.
        n (Optional[int]): The number of completion choices to return. Currently, only 1 choice is supported.
    """
    messages: List[dict]
    user: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    stream: Optional[bool] = None
    stop: Optional[Union[str, List[str]]] = None
    n: Optional[int] = None


class ChatCompletion(FoundationModelAPIResource):
    """
    A class representing the chat completion API resource.
    """
    SUPPORTED_MODEL_LIST = [model.value for model in ChatCompletionModel.__members__.values()]
    model_input = ChatCompletionAPIInput
    model_output = ChatCompletionObject
    model_streaming_output = ChatCompletionChunkObject

    @classmethod
    def create(cls, **kwargs):
        return super().create(**kwargs)
